import numpy as np


def find_rigid_transform(A: np.ndarray, B: np.ndarray, force_reflection: bool = False):
    """Calculate rigid transformation to map coordinates A onto coordinates B.

    Parameters
    ----------

    A: np.ndarray
        Array of 2D or 3D source coordinates with shape [2xN] or [3xN]

    B: np.ndarray
        Array of 2D or 3D target coordinates A with shape [2xN] or [3xN]

    force_reflection: bool
        Whether to force reflection if one is suspected: it can give false positives (optional, default False).


    Returns
    -------

    R: np.ndarray
        Rotation matrix (either 2x2 or 3x3) to transform the A coordinates into B.

    t: np.ndarray
        Translation vector (either 2x1 or 3x1) to transform the A coordinates into B.
    """
    assert A.shape == B.shape

    num_rows, num_cols = A.shape
    if num_rows not in [2, 3]:
        raise ValueError(f"matrix A is not 2xN or 3xN, it is {num_rows}x{num_cols}")

    num_rows, num_cols = B.shape
    if num_rows not in [2, 3]:
        raise ValueError(f"matrix B is not 2xN or 3xN, it is {num_rows}x{num_cols}")

    # Number of dimensions
    nDim = num_rows

    # Find column mean
    centroid_A = np.mean(A, axis=1)
    centroid_B = np.mean(B, axis=1)

    # Ensure centroids are column vectors
    centroid_A = centroid_A.reshape(-1, 1)
    centroid_B = centroid_B.reshape(-1, 1)

    # Subtract mean
    Am = A - centroid_A
    Bm = B - centroid_B

    H = Am @ np.transpose(Bm)

    # Sanity check
    rank_H = np.linalg.matrix_rank(H)
    if rank_H < nDim:
        raise ValueError(f"Rank of H = {rank_H}; expected {nDim}.")

    # Find rotation matrix R
    U, S, Vt = np.linalg.svd(H)
    R = Vt.T @ U.T

    # Special reflection case
    if np.linalg.det(R) < 0:
        if force_reflection:
            print("det(R) < R, reflection detected! Correcting...")
            Vt[-1, :] *= -1
            R = Vt.T @ U.T
        else:
            print("det(R) < R, reflection detected! As requested, not correcting...")

    # Calculate the translation vector
    t = -R @ centroid_A + centroid_B

    # Return transposed versions to work on [NxnDim] arrays
    return R, t


def apply_rigid_transform(A: np.ndarray, R: np.ndarray, t: np.ndarray):
    """Applies rotation and translation to coordinates A.

    Parameters
    ----------

    A: np.ndarray
        Array of 2D or 3D source coordinates with shape [2xN] or [3xN]

    R: np.ndarray
        Rotation matrix (2x2 or 3x3).

    t: np.ndarray
        Translation vector (2x1 or 3x1)

    Returns
    -------

    A_t: np.ndarray
        Array of 2D or 3D coordinates (shape [2xN] or [3xN]) transformed by rotation matrix R and translation vector t.
    """

    # Check dimensions
    num_rows, num_cols = A.shape
    if num_rows not in [2, 3]:
        raise ValueError(f"Matrix A is not 2xN or 3xN, it is {num_rows}x{num_cols}")

    if R.shape[0] != R.shape[1] and R.shape[0] != num_rows:
        raise ValueError(f"Rotation matrix R must be {num_cols}x{num_cols}")

    if t.shape[0] != num_rows:
        raise ValueError(f"Translation vector t must be {num_rows}x1")

    # Rotate and translate
    return (R @ A) + t
